@@ -1,5 +1,5 @@
 /*
- * $Id: apachelib.c 93 2005-08-03 01:32:41Z shugo $
+ * $Id: apachelib.c 135 2008-01-15 16:20:18Z shugo $
  * Copyright (C) 2000  ZetaBITS, Inc.
  * Copyright (C) 2000  Information-technology Promotion Agency, Japan
  * Copyright (C) 2001-2003  Shugo Maeda <shugo@modruby.net>
@@ -131,9 +131,9 @@ static VALUE apache_unescape_url(VALUE self, VALUE url)
     char *buf;
 
     Check_Type(url, T_STRING);
-    buf = ALLOCA_N(char, RSTRING(url)->len + 1);
-    memcpy(buf, RSTRING(url)->ptr, RSTRING(url)->len);
-    buf[RSTRING(url)->len] = '\0';
+    buf = ALLOCA_N(char, RSTRING_LEN(url) + 1);
+    memcpy(buf, RSTRING_PTR(url), RSTRING_LEN(url));
+    buf[RSTRING_LEN(url)] = '\0';
     ap_unescape_url(buf);
     return rb_str_new2(buf);
 }
@@ -389,6 +389,32 @@ void rb_init_apache()
     rb_define_const(rb_mApache, "REMOTE_DOUBLE_REV",
 		    INT2NUM(REMOTE_DOUBLE_REV));
 
+#ifdef APACHE2
+    rb_define_const(rb_mApache, "AP_CONN_UNKNOWN",
+		    INT2NUM(AP_CONN_UNKNOWN));
+    rb_define_const(rb_mApache, "AP_CONN_CLOSE",
+		    INT2NUM(AP_CONN_CLOSE));
+    rb_define_const(rb_mApache, "AP_CONN_KEEPALIVE",
+		    INT2NUM(AP_CONN_KEEPALIVE));
+#endif
+
+    rb_define_const(rb_mApache, "APLOG_EMERG",
+		    INT2NUM(APLOG_EMERG));
+    rb_define_const(rb_mApache, "APLOG_ALERT",
+		    INT2NUM(APLOG_ALERT));
+    rb_define_const(rb_mApache, "APLOG_CRIT",
+		    INT2NUM(APLOG_CRIT));
+    rb_define_const(rb_mApache, "APLOG_ERR",
+		    INT2NUM(APLOG_ERR));
+    rb_define_const(rb_mApache, "APLOG_WARNING",
+		    INT2NUM(APLOG_WARNING));
+    rb_define_const(rb_mApache, "APLOG_NOTICE",
+		    INT2NUM(APLOG_NOTICE));
+    rb_define_const(rb_mApache, "APLOG_INFO",
+		    INT2NUM(APLOG_INFO));
+    rb_define_const(rb_mApache, "APLOG_DEBUG",
+		    INT2NUM(APLOG_DEBUG));
+
     /* Policy constants for setup_client_block() */
     rb_define_const(rb_mApache, "REQUEST_NO_BODY",
 		    INT2NUM(REQUEST_NO_BODY));
@@ -425,6 +451,8 @@ void rb_init_apache()
 #ifdef APACHE2
     rb_init_apache_bucket();
 #endif
+    rb_init_apache_uri();
+    rb_init_apache_error();
 }
 
 /*
