@@ -1,5 +1,5 @@
 /*
- * $Id: mod_ruby.c 126 2006-05-27 03:10:03Z shugo $
+ * $Id: mod_ruby.c 144 2008-11-08 04:34:48Z shugo $
  * Copyright (C) 2000  ZetaBITS, Inc.
  * Copyright (C) 2000  Information-technology Promotion Agency, Japan
  * Copyright (C) 2001  Shugo Maeda <shugo@modruby.net>
@@ -81,7 +81,6 @@ extern char **environ;
 #define FREE_ENVIRON(e)
 #endif
 
-RUBY_EXTERN VALUE ruby_errinfo;
 RUBY_EXTERN VALUE rb_stdin;
 RUBY_EXTERN VALUE rb_stdout;
 RUBY_EXTERN VALUE rb_stderr;
@@ -95,10 +94,9 @@ static VALUE orig_stdout;
 static VALUE orig_defout;
 #endif
 
-RUBY_EXTERN VALUE rb_load_path;
 static VALUE default_load_path;
 
-RUBY_EXTERN VALUE rb_progname;
+static VALUE progname;
 
 static const char *default_kcode;
 
@@ -198,6 +196,10 @@ static const command_rec ruby_cmds[] = {
      "set fixup handler object"),
     AP_INIT_TAKE1("RubyLogHandler", ruby_cmd_log_handler, NULL, OR_ALL,
      "set log handler object"),
+#ifdef APACHE2
+    AP_INIT_TAKE1("RubyErrorLogHandler", ruby_cmd_error_log_handler, NULL, OR_ALL,
+     "set log handler object"),
+#endif
     AP_INIT_TAKE1("RubyHeaderParserHandler", ruby_cmd_header_parser_handler,
      NULL, OR_ALL,
      "set header parser object"),
@@ -220,6 +222,7 @@ static const command_rec ruby_cmds[] = {
 
 static int ruby_startup(pool*, pool*, pool*, server_rec*);
 static void ruby_child_init(pool*, server_rec*);
+static void ruby_error_log_handler(const char*, int, int, apr_status_t, const server_rec*, const request_rec*, pool*, const char*);
 
 static void ruby_register_hooks(pool *p)
 {
@@ -233,6 +236,9 @@ static void ruby_register_hooks(pool *p)
     ap_hook_type_checker(ruby_type_handler, NULL, NULL, APR_HOOK_MIDDLE);
     ap_hook_fixups(ruby_fixup_handler, NULL, NULL, APR_HOOK_MIDDLE);
     ap_hook_log_transaction(ruby_log_handler, NULL, NULL, APR_HOOK_MIDDLE);
+#ifdef APACHE2
+    ap_hook_error_log(ruby_error_log_handler, NULL, NULL, APR_HOOK_MIDDLE);
+#endif
     ap_hook_child_init(ruby_child_init, NULL, NULL, APR_HOOK_MIDDLE);
     ap_hook_post_read_request(ruby_post_read_request_handler,
 			      NULL, NULL, APR_HOOK_MIDDLE);
@@ -341,13 +347,13 @@ static void get_error_pos(VALUE str)
     char buff[BUFSIZ];
     ID this_func = rb_frame_this_func();
 
-    if (ruby_sourcefile) {
+    if (rb_sourcefile()) {
 	if (this_func) {
-	    snprintf(buff, BUFSIZ, "%s:%d:in `%s'", ruby_sourcefile, ruby_sourceline,
+	    snprintf(buff, BUFSIZ, "%s:%d:in `%s'", rb_sourcefile(), rb_sourceline(),
 		     rb_id2name(this_func));
 	}
 	else {
-	    snprintf(buff, BUFSIZ, "%s:%d", ruby_sourcefile, ruby_sourceline);
+	    snprintf(buff, BUFSIZ, "%s:%d", rb_sourcefile(), rb_sourceline());
 	}
 	rb_str_cat(str, buff, strlen(buff));
     }
@@ -362,29 +368,29 @@ static void get_exception_info(VALUE str)
     int elen;
     int state;
 
-    if (NIL_P(ruby_errinfo)) return;
+    if (NIL_P(rb_errinfo())) return;
 
-    errat = rb_funcall(ruby_errinfo, rb_intern("backtrace"), 0);
+    errat = rb_funcall(rb_errinfo(), rb_intern("backtrace"), 0);
     if (!NIL_P(errat)) {
-	VALUE mesg = RARRAY(errat)->ptr[0];
+	VALUE mesg = RARRAY_PTR(errat)[0];
 
 	if (NIL_P(mesg)) {
 	    get_error_pos(str);
 	}
 	else {
-	    rb_str_cat(str, RSTRING(mesg)->ptr, RSTRING(mesg)->len);
+	    rb_str_cat(str, RSTRING_PTR(mesg), RSTRING_LEN(mesg));
 	}
     }
 
-    eclass = CLASS_OF(ruby_errinfo);
-    estr = rb_protect(rb_obj_as_string, ruby_errinfo, &state);
+    eclass = CLASS_OF(rb_errinfo());
+    estr = rb_protect(rb_obj_as_string, rb_errinfo(), &state);
     if (state) {
 	einfo = "";
 	elen = 0;
     }
     else {
-	einfo = RSTRING(estr)->ptr;
-	elen = RSTRING(estr)->len;
+	einfo = RSTRING_PTR(estr);
+	elen = RSTRING_LEN(estr);
     }
     if (eclass == rb_eRuntimeError && elen == 0) {
 	STR_CAT_LITERAL(str, ": unhandled exception\n");
@@ -395,14 +401,14 @@ static void get_exception_info(VALUE str)
 	epath = rb_class_path(eclass);
 	if (elen == 0) {
 	    STR_CAT_LITERAL(str, ": ");
-	    rb_str_cat(str, RSTRING(epath)->ptr, RSTRING(epath)->len);
+	    rb_str_cat(str, RSTRING_PTR(epath), RSTRING_LEN(epath));
 	    STR_CAT_LITERAL(str, "\n");
 	}
 	else {
 	    char *tail  = 0;
 	    int len = elen;
 
-	    if (RSTRING(epath)->ptr[0] == '#') epath = 0;
+	    if (RSTRING_PTR(epath)[0] == '#') epath = 0;
 	    if ((tail = strchr(einfo, '\n')) != NULL) {
 		len = tail - einfo;
 		tail++;		/* skip newline */
@@ -411,7 +417,7 @@ static void get_exception_info(VALUE str)
 	    rb_str_cat(str, einfo, len);
 	    if (epath) {
 		STR_CAT_LITERAL(str, " (");
-		rb_str_cat(str, RSTRING(epath)->ptr, RSTRING(epath)->len);
+		rb_str_cat(str, RSTRING_PTR(epath), RSTRING_LEN(epath));
 		STR_CAT_LITERAL(str, ")\n");
 	    }
 	    if (tail) {
@@ -423,18 +429,18 @@ static void get_exception_info(VALUE str)
 
     if (!NIL_P(errat)) {
 	long i, len;
-	struct RArray *ep;
+	VALUE *ep;
 
 #define TRACE_MAX (TRACE_HEAD+TRACE_TAIL+5)
 #define TRACE_HEAD 8
 #define TRACE_TAIL 5
 
-	ep = RARRAY(errat);
-	len = ep->len;
+	ep = RARRAY_PTR(errat);
+	len = RARRAY_LEN(errat);
 	for (i=1; i<len; i++) {
-	    if (TYPE(ep->ptr[i]) == T_STRING) {
+	    if (TYPE(ep[i]) == T_STRING) {
 		STR_CAT_LITERAL(str, "  from ");
-		rb_str_cat(str, RSTRING(ep->ptr[i])->ptr, RSTRING(ep->ptr[i])->len);
+		rb_str_cat(str, RSTRING_PTR(ep[i]), RSTRING_LEN(ep[i]));
 		STR_CAT_LITERAL(str, "\n");
 	    }
 	    if (i == TRACE_HEAD && len > TRACE_MAX) {
@@ -446,7 +452,7 @@ static void get_exception_info(VALUE str)
 	    }
 	}
     }
-    /* ruby_errinfo = Qnil; */
+    /* rb_errinfo() = Qnil; */
 }
 
 VALUE ruby_get_error_info(int state)
@@ -512,9 +518,9 @@ void ruby_log_error_string(server_rec *s, VALUE errmsg)
 
     ruby_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s, "error in ruby");
     msgs = rb_str_split(errmsg, "\n");
-    for (i = 0; i < RARRAY(msgs)->len; i++) {
+    for (i = 0; i < RARRAY_LEN(msgs); i++) {
         ruby_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s,
-                       "%s", StringValuePtr(RARRAY(msgs)->ptr[i]));
+                       "%s", StringValuePtr(RARRAY_PTR(msgs)[i]));
     }
 }
 
@@ -528,7 +534,7 @@ static void handle_error(request_rec *r, int state)
 	rconf = get_request_config(r);
 	if (rconf && !NIL_P(rconf->request_object))
 	    rb_apache_request_set_error(rconf->request_object,
-					errmsg, ruby_errinfo);
+					errmsg, rb_errinfo());
     }
     ruby_log_error_string(r->server, errmsg);
 }
@@ -543,23 +549,29 @@ void mod_ruby_setup_loadpath(ruby_server_config *sconf,
 {
     int i, n;
     char **paths;
+    VALUE load_path;
 
-    rb_load_path = rb_ary_new();
-    for (i = 0; i < RARRAY(default_load_path)->len; i++) {
-	rb_ary_push(rb_load_path, rb_str_dup(RARRAY(default_load_path)->ptr[i]));
+#ifdef RUBY_VM
+    load_path = GET_LOAD_PATH();
+    rb_ary_clear(load_path);
+#else
+    rb_load_path = load_path = rb_ary_new();
+#endif
+    for (i = 0; i < RARRAY_LEN(default_load_path); i++) {
+	rb_ary_push(load_path, rb_str_dup(RARRAY_PTR(default_load_path)[i]));
     }
     if (sconf && sconf->load_path) {
 	paths = (char **) sconf->load_path->elts;
 	n = sconf->load_path->nelts;
 	for (i = 0; i < n; i++) {
-	    rb_ary_push(rb_load_path, rb_str_new2(paths[i]));
+	    rb_ary_push(load_path, rb_str_new2(paths[i]));
 	}
     }
     if (dconf && dconf->load_path) {
 	paths = (char **) dconf->load_path->elts;
 	n = dconf->load_path->nelts;
 	for (i = 0; i < n; i++) {
-	    rb_ary_push(rb_load_path, rb_str_new2(paths[i]));
+	    rb_ary_push(load_path, rb_str_new2(paths[i]));
 	}
     }
 }
@@ -575,8 +587,8 @@ static int ruby_require_directly(const char *filename,
     fname = rb_str_new2(filename);
     rb_protect_funcall(Qnil, rb_intern("require"), &state, 1, fname);
     if (state == TAG_RAISE &&
-	rb_obj_is_kind_of(ruby_errinfo, rb_eSystemExit)) {
-	exit_status = rb_iv_get(ruby_errinfo, "status");
+	rb_obj_is_kind_of(rb_errinfo(), rb_eSystemExit)) {
+	exit_status = rb_iv_get(rb_errinfo(), "status");
 	exit(NUM2INT(exit_status));
     }
     return state;
@@ -631,7 +643,6 @@ static void ruby_startup(server_rec *s, pool *p)
 
 static void ruby_init_interpreter(server_rec *s)
 {
-    VALUE stack_start;
     ruby_server_config *conf = get_server_config(s);
     ruby_library_context *libraries;
     char **list;
@@ -646,8 +657,15 @@ static void ruby_init_interpreter(server_rec *s)
 #ifdef SIGTERM
     RETSIGTYPE (*sigterm_handler)_((int));
 #endif
+#ifdef RUBY_VM
+    void Init_prelude(void);
+    RUBY_INIT_STACK;
+#else
+    VALUE stack_start;
     void Init_stack _((VALUE*));
 
+    Init_stack(&stack_start);
+#endif
 #ifdef SIGHUP
     sighup_handler = signal(SIGHUP, SIG_DFL);
 #endif
@@ -670,8 +688,6 @@ static void ruby_init_interpreter(server_rec *s)
     if (sigterm_handler != SIG_ERR)
 	ruby_signal(SIGTERM, sigterm_handler);
 #endif
-
-    Init_stack(&stack_start);
     rb_init_apache();
 
     rb_define_global_const("MOD_RUBY",
@@ -687,8 +703,12 @@ static void ruby_init_interpreter(server_rec *s)
 #endif
 
     ruby_init_loadpath();
-    default_load_path = rb_load_path;
+#ifdef RUBY_VM
+    Init_prelude();
+#endif
+    default_load_path = GET_LOAD_PATH();
     rb_global_variable(&default_load_path);
+    rb_define_variable("$0", &progname);
     list = (char **) conf->load_path->elts;
     n = conf->load_path->nelts;
     for (i = 0; i < n; i++) {
@@ -733,8 +753,8 @@ static void ruby_finalize_interpreter()
     int i;
 
     ruby_finalize();
-    for (i = 0; i < RARRAY(ruby_dln_librefs)->len; i++) {
-	dso_unload((void *) NUM2LONG(RARRAY(ruby_dln_librefs)->ptr[i]));
+    for (i = 0; i < RARRAY_LEN(ruby_dln_librefs); i++) {
+	dso_unload((void *) NUM2LONG(RARRAY_PTR(ruby_dln_librefs)[i]));
     }
 }
 
@@ -851,7 +871,7 @@ static request_rec *fake_request_rec(server_rec *s, pool *p, char *hook)
     return r;
 }
 
-static int ruby_handler(request_rec *, array_header *, ID, int, int);
+static int ruby_handler(request_rec *, array_header *, error_log_data *, ID, int, int);
 
 #ifdef APACHE2
 static void ruby_child_init(pool *p, server_rec *s)
@@ -916,7 +936,7 @@ static void ruby_child_init(server_rec *s, pool *p)
 
     r = fake_request_rec(s, p, "RubyChildInitHandler");
     conf = get_server_config(r->server);
-    ruby_handler(r, conf->ruby_child_init_handler,
+    ruby_handler(r, conf->ruby_child_init_handler, NULL,
 		 rb_intern("child_init"), 0, 0);
 }
 
@@ -992,8 +1012,8 @@ static VALUE kill_threads(VALUE arg)
     VALUE th;
     int i;
 
-    for (i = 0; i < RARRAY(threads)->len; i++) {
-	th = RARRAY(threads)->ptr[i];
+    for (i = 0; i < RARRAY_LEN(threads); i++) {
+	th = RARRAY_PTR(threads)[i];
 	if (th != main_thread)
 	    rb_protect_funcall(th, rb_intern("kill"), NULL, 0);
     }
@@ -1032,11 +1052,13 @@ static VALUE run_safely_0(VALUE arg)
     VALUE result;
 
     rb_set_safe_level(rsarg->safe_level);
+#ifndef RUBY_VM /* TODO */
     if (rsarg->timeout > 0) {
 	targ.thread = rb_thread_current();
 	targ.timeout = rsarg->timeout;
 	timeout_thread = rb_thread_create(do_timeout, (void *) &targ);
     }
+#endif
     result = (*rsarg->func)(rsarg->arg);
     if (!NIL_P(timeout_thread))
 	rb_protect_funcall(timeout_thread, rb_intern("kill"), NULL, 0);
@@ -1054,18 +1076,22 @@ static int run_safely(int safe_level, int timeout,
     rsarg.timeout = timeout;
     rsarg.func = func;
     rsarg.arg = arg;
-#if defined(HAVE_SETITIMER)
+#if !defined(RUBY_VM) && defined(HAVE_SETITIMER)
     rb_thread_start_timer();
 #endif
-    if (safe_level > ruby_safe_level) {
+#ifndef RUBY_VM /* TODO */
+    if (safe_level > rb_safe_level()) {
 	thread = rb_thread_create(run_safely_0, &rsarg);
 	ret = rb_protect_funcall(thread, rb_intern("value"), &state, 0);
     }
     else {
+#endif
 	ret = rb_protect(run_safely_0, (VALUE) &rsarg, &state);
+#ifndef RUBY_VM /* TODO */
     }
+#endif
     rb_protect(kill_threads, Qnil, NULL);
-#if defined(HAVE_SETITIMER)
+#if !defined(RUBY_VM) && defined(HAVE_SETITIMER)
     rb_thread_stop_timer();
 #endif
     if (retval)
@@ -1115,10 +1141,14 @@ static void per_request_init(request_rec *r)
     ruby_dir_config *dconf;
 
     if (r->request_config) {
-	rconf = apr_palloc(r->pool, sizeof(ruby_request_config));
-	rconf->saved_env = save_env(r->pool);
-	rconf->request_object = Qnil;
-	ap_set_module_config(r->request_config, &ruby_module, rconf);
+	rconf = get_request_config(r);
+        /* may already have request object from successive handler calls */
+	if (rconf == NULL || NIL_P(rconf->request_object)) {
+            rconf = apr_palloc(r->pool, sizeof(ruby_request_config));
+            rconf->saved_env = save_env(r->pool);
+            rconf->request_object = Qnil;
+            ap_set_module_config(r->request_config, &ruby_module, rconf);
+        }
     }
     dconf = get_dir_config(r);
     sconf = get_server_config(r->server);
@@ -1133,7 +1163,7 @@ static void per_request_init(request_rec *r)
     rb_defout = rb_request;
 #endif
     if (r->filename)
-	rb_progname = rb_tainted_str_new2(r->filename);
+	progname = rb_tainted_str_new2(r->filename);
 }
 
 static VALUE exec_end_proc(VALUE arg)
@@ -1176,7 +1206,7 @@ static void per_request_cleanup(request_rec *r, int flush)
 	if (rconf)
 	    restore_env(r->pool, rconf->saved_env);
     }
-    rb_progname = Qnil;
+    progname = Qnil;
     if (dconf && dconf->gc_per_request)
 	rb_gc();
 }
@@ -1184,6 +1214,7 @@ static void per_request_cleanup(request_rec *r, int flush)
 typedef struct handler_0_arg {
     request_rec *r;
     char *handler;
+    error_log_data *err;
     ID mid;
 } handler_0_arg_t;
 
@@ -1191,17 +1222,23 @@ static VALUE ruby_handler_0(void *arg)
 {
     handler_0_arg_t *ha = (handler_0_arg_t *) arg;
     request_rec *r = ha->r;
+    error_log_data *err = ha->err;
     char *handler = ha->handler;
     ID mid = ha->mid;
     VALUE ret;
     int state;
 
-    ret = rb_protect_funcall(rb_eval_string(handler), mid, &state,
-			     1, rb_request);
+    if (err) {
+        ret = rb_protect_funcall(rb_eval_string(handler), mid, &state,
+		                 2, rb_request, rb_apache_error_new(r, err));
+    } else {
+        ret = rb_protect_funcall(rb_eval_string(handler), mid, &state,
+			         1, rb_request);
+    }
     if (state) {
 	if (state == TAG_RAISE &&
-	    rb_obj_is_kind_of(ruby_errinfo, rb_eSystemExit)) {
-	    ret = rb_iv_get(ruby_errinfo, "status");
+	    rb_obj_is_kind_of(rb_errinfo(), rb_eSystemExit)) {
+	    ret = rb_iv_get(rb_errinfo(), "status");
 	}
 	else {
 	    handle_error(r, state);
@@ -1222,6 +1259,7 @@ static VALUE ruby_handler_0(void *arg)
 typedef struct handler_internal_arg {
     request_rec *r;
     array_header *handlers_arr;
+    error_log_data *err;
     ID mid;
     int run_all;
     int flush;
@@ -1232,6 +1270,7 @@ static void *ruby_handler_internal(handler_internal_arg_t *iarg)
 {
     request_rec *r = iarg->r;
     array_header *handlers_arr = iarg->handlers_arr;
+    error_log_data *err = iarg->err;
     ID mid = iarg->mid;
     int run_all = iarg->run_all;
     int flush = iarg->flush;
@@ -1259,6 +1298,7 @@ static void *ruby_handler_internal(handler_internal_arg_t *iarg)
     for (i = 0; i < handlers_len; i++) {
 	arg.r = r;
 	arg.handler = handlers[i];
+	arg.err = err;
 	arg.mid = mid;
 	ap_soft_timeout("call ruby handler", r);
 	timeout = sconf->timeout;
@@ -1289,8 +1329,9 @@ static void *ruby_handler_internal(handler_internal_arg_t *iarg)
 }
 
 static int ruby_handler(request_rec *r,
-			array_header *handlers_arr, ID mid,
-			int run_all, int flush)
+			array_header *handlers_arr,
+			error_log_data *error,
+			ID mid, int run_all, int flush)
 {
     handler_internal_arg_t *arg;
 
@@ -1300,6 +1341,7 @@ static int ruby_handler(request_rec *r,
     arg = apr_palloc(r->pool, sizeof(handler_internal_arg_t));
     arg->r = r;
     arg->handlers_arr = handlers_arr;
+    arg->err = error;
     arg->mid = mid;
     arg->run_all = run_all;
     arg->flush = flush;
@@ -1340,7 +1382,7 @@ static int ruby_object_handler(request_rec *r)
     }
 #endif
     dconf = get_dir_config(r);
-    retval = ruby_handler(r, dconf->ruby_handler, rb_intern("handler"), 0, 1);
+    retval = ruby_handler(r, dconf->ruby_handler, NULL, rb_intern("handler"), 0, 1);
 #ifdef APACHE2
     if (retval == DECLINED && r->finfo.filetype == APR_DIR)
         r->handler = DIR_MAGIC_TYPE;
@@ -1352,7 +1394,7 @@ static int ruby_trans_handler(request_rec *r)
 {
     ruby_dir_config *dconf = get_dir_config(r);
 
-    return ruby_handler(r, dconf->ruby_trans_handler,
+    return ruby_handler(r, dconf->ruby_trans_handler, NULL,
 			rb_intern("translate_uri"), 0, 0);
 }
 
@@ -1362,7 +1404,7 @@ static int ruby_authen_handler(request_rec *r)
     int retval;
 
     if (dconf->ruby_authen_handler == NULL) return DECLINED;
-    retval = ruby_handler(r, dconf->ruby_authen_handler,
+    retval = ruby_handler(r, dconf->ruby_authen_handler, NULL,
 			  rb_intern("authenticate"), 0, 0);
     return retval;
 }
@@ -1371,7 +1413,7 @@ static int ruby_authz_handler(request_rec *r)
 {
     ruby_dir_config *dconf = get_dir_config(r);
 
-    return ruby_handler(r, dconf->ruby_authz_handler,
+    return ruby_handler(r, dconf->ruby_authz_handler, NULL,
 			rb_intern("authorize"), 0, 0);
 }
 
@@ -1379,7 +1421,7 @@ static int ruby_access_handler(request_rec *r)
 {
     ruby_dir_config *dconf = get_dir_config(r);
 
-    return ruby_handler(r, dconf->ruby_access_handler,
+    return ruby_handler(r, dconf->ruby_access_handler, NULL,
 			rb_intern("check_access"), 1, 0);
 }
 
@@ -1388,7 +1430,7 @@ static int ruby_type_handler(request_rec *r)
     ruby_dir_config *dconf = get_dir_config(r);
 
     if (dconf->ruby_type_handler == NULL) return DECLINED;
-    return ruby_handler(r, dconf->ruby_type_handler,
+    return ruby_handler(r, dconf->ruby_type_handler, NULL,
 			rb_intern("find_types"), 0, 0);
 }
 
@@ -1397,7 +1439,7 @@ static int ruby_fixup_handler(request_rec *r)
     ruby_dir_config *dconf = get_dir_config(r);
 
     if (dconf->ruby_fixup_handler == NULL) return DECLINED;
-    return ruby_handler(r, dconf->ruby_fixup_handler,
+    return ruby_handler(r, dconf->ruby_fixup_handler, NULL,
 			rb_intern("fixup"), 1, 0);
 }
 
@@ -1406,7 +1448,7 @@ static int ruby_log_handler(request_rec *r)
     ruby_dir_config *dconf = get_dir_config(r);
 
     if (dconf->ruby_log_handler == NULL) return DECLINED;
-    return ruby_handler(r, dconf->ruby_log_handler,
+    return ruby_handler(r, dconf->ruby_log_handler, NULL,
 			rb_intern("log_transaction"), 1, 0);
 }
 
@@ -1418,13 +1460,13 @@ static int ruby_header_parser_handler(request_rec *r)
 
     if (dconf->ruby_init_handler &&
 	ap_table_get(r->notes, "ruby_init_ran") == NULL) {
-	retval = ruby_handler(r, dconf->ruby_init_handler,
+	retval = ruby_handler(r, dconf->ruby_init_handler, NULL,
 			      rb_intern("init"), 1, 0);
 	if (retval != OK && retval != DECLINED)
 	    return retval;
     }
     if (dconf->ruby_header_parser_handler == NULL) return DECLINED;
-    return ruby_handler(r, dconf->ruby_header_parser_handler,
+    return ruby_handler(r, dconf->ruby_header_parser_handler, NULL,
 			rb_intern("header_parse"), 1, 0);
 }
 #endif
@@ -1434,7 +1476,7 @@ static APR_CLEANUP_RETURN_TYPE ruby_cleanup_handler(void *data)
     request_rec *r = (request_rec *) data;
     ruby_dir_config *dconf = get_dir_config(r);
 
-    ruby_handler(r, dconf->ruby_cleanup_handler,
+    ruby_handler(r, dconf->ruby_cleanup_handler, NULL,
 		 rb_intern("cleanup"), 1, 0);
     APR_CLEANUP_RETURN_SUCCESS();
 }
@@ -1448,16 +1490,43 @@ static int ruby_post_read_request_handler(request_rec *r)
 			apr_pool_cleanup_null);
 
     if (dconf->ruby_init_handler) {
-	retval = ruby_handler(r, dconf->ruby_init_handler,
+	retval = ruby_handler(r, dconf->ruby_init_handler, NULL,
 			      rb_intern("init"), 1, 0);
 	apr_table_set(r->notes, "ruby_init_ran", "true");
 	if (retval != OK && retval != DECLINED)
 	    return retval;
     }
-    return ruby_handler(r, dconf->ruby_post_read_request_handler,
+    return ruby_handler(r, dconf->ruby_post_read_request_handler, NULL,
 			rb_intern("post_read_request"), 1, 0);
 }
 
+#ifdef APACHE2
+static void ruby_error_log_handler(const char *file, int line, int level, apr_status_t status, const server_rec *s, const request_rec *r, apr_pool_t *pool, const char *error) {
+    ruby_dir_config *dconf;
+    error_log_data *e;
+ 
+    /* only interested in errors produced as the result of requests.
+     * 
+     * [this handler is called multiple times during startup but since this is
+     * prior to config stage, we'll ignore them] */
+    if (r == NULL) return;
+
+    dconf = get_dir_config(r);
+    if (dconf->ruby_error_log_handler == NULL) return;
+    
+    e = apr_palloc(r->pool, sizeof(error_log_data));
+    e->file = file;
+    e->line = line;
+    e->level = level;
+    e->status = status;
+    e->error = error;
+
+    ruby_handler((request_rec *) r, dconf->ruby_error_log_handler, e,
+                 rb_intern("log_error"), 1, 0);
+    return;
+}
+#endif
+
 /*
  * Local variables:
  * mode: C
