@@ -1,6 +1,6 @@
 /*
  * multival.c - The Apache::MultiVal class
- * $Id: multival.c 16 2004-07-16 09:16:29Z shugo $
+ * $Id: multival.c 142 2008-11-08 03:57:19Z shugo $
  *
  * Author: Michael Granger <mgranger@RubyCrafters.com>
  * Copyright (c) 2003 RubyCrafters, LLC. All rights reserved.
@@ -129,17 +129,17 @@ multival_init( VALUE self, VALUE args )
     long len, i;
 
     /* Make sure there's at least one argument to work with */
-    if ( RARRAY(args)->len == 0 )
+    if ( RARRAY_LEN(args) == 0 )
         rb_ary_push( args, rb_tainted_str_new("", 0) );
 
     /* Stringify all arguments */
-    len = RARRAY(args)->len;
+    len = RARRAY_LEN(args);
     collect = rb_ary_new2(len);
     for ( i = 0; i < len; i++ ) {
         VALUE str;
 		
-        str = rb_str_dup( rb_obj_as_string(RARRAY(args)->ptr[i]) );
-        OBJ_INFECT( str, RARRAY(args)->ptr[i] );
+        str = rb_str_dup( rb_obj_as_string(RARRAY_PTR(args)[i]) );
+        OBJ_INFECT( str, RARRAY_PTR(args)[i] );
 		
         rb_ary_push( collect, str );
     }
@@ -198,7 +198,7 @@ multival_compare( VALUE self, VALUE other )
 }
 
 
-
+#if 0
 /*
  * Delegate the current method to the stringish version of the multival (ie.,
  * the first element of the array).
@@ -206,10 +206,10 @@ multival_compare( VALUE self, VALUE other )
 static VALUE
 multival_string_delegator( int argc, VALUE *argv, VALUE self )
 {
-    ID meth = rb_frame_last_func();
+    ID meth = rb_frame_this_func();
     VALUE args = rb_iv_get( self, "@args" );
 	
-    return rb_funcall2( RARRAY(args)->ptr[0], meth, argc, argv );
+    return rb_funcall2( RARRAY_PTR(args)[0], meth, argc, argv );
 }
 
 
@@ -219,7 +219,7 @@ multival_string_delegator( int argc, VALUE *argv, VALUE self )
 static VALUE
 multival_array_delegator( int argc, VALUE *argv, VALUE self )
 {
-    ID meth = rb_frame_last_func();
+    ID meth = rb_frame_this_func();
     VALUE args = rb_iv_get( self, "@args" );
 	
     return rb_funcall2( args, meth, argc, argv );
@@ -241,6 +241,7 @@ multival_make_delegator( VALUE name, ID which )
 	
     return Qtrue;
 }
+#endif
 
 
 /* Module initializer */
@@ -269,11 +270,14 @@ void rb_init_apache_multival()
     rb_ary_delete( dmethods, rb_str_new2("each") );
     rb_ary_delete( dmethods, rb_str_new2("[]") );
     rb_ary_delete( dmethods, rb_str_new2("[]=") );
+    /* TODO: avoid SEGV */
+#if 0
     rb_iterate( rb_each, dmethods, multival_make_delegator, stringish );
 
     /* Now set up array-ish delegation */	
     dmethods = rb_class_instance_methods( 1, args, rb_cArray );
     rb_iterate( rb_each, dmethods, multival_make_delegator, arrayish );
+#endif
 	
     /* include Comparable */
     rb_include_module( rb_cApacheMultiVal, rb_mComparable );
