@@ -1,6 +1,6 @@
 /*
  * paramtable.c - The Apache::ParamTable class
- * $Id: paramtable.c 104 2006-01-31 17:52:04Z shugo $
+ * $Id: paramtable.c 142 2008-11-08 03:57:19Z shugo $
  *
  * Author: Michael Granger <mgranger@RubyCrafters.com>
  * Copyright (c) 2003 RubyCrafters, LLC. All rights reserved.
@@ -150,8 +150,8 @@ static VALUE paramtable_set( VALUE self, VALUE name, VALUE val )
         int i;
 		
         apr_table_unset( tbl, key );
-        for ( i = 0; i < RARRAY(ary)->len; i++ ) {
-            str = rb_check_convert_type( *(RARRAY(ary)->ptr+i), T_STRING,
+        for ( i = 0; i < RARRAY_LEN(ary); i++ ) {
+            str = rb_check_convert_type( *(RARRAY_PTR(ary)+i), T_STRING,
                                          "String", "to_str" );
             apr_table_add( tbl, key, StringValuePtr(str) );
         }
