@@ -1,5 +1,5 @@
 /*
- * $Id: request.c 104 2006-01-31 17:52:04Z shugo $
+ * $Id: request.c 142 2008-11-08 03:57:19Z shugo $
  * Copyright (C) 2000  ZetaBITS, Inc.
  * Copyright (C) 2000  Information-technology Promotion Agency, Japan
  * Copyright (C) 2001-2003  Shugo Maeda <shugo@modruby.net>
@@ -36,7 +36,11 @@
 #include "mod_ruby.h"
 #include "apachelib.h"
 
+#ifdef RUBY_VM
+#include "ruby/st.h"
+#else
 #include "st.h"
+#endif
 
 static VALUE rb_eApachePrematureChunkEndError;
 VALUE rb_cApacheRequest;
@@ -56,6 +60,7 @@ typedef struct request_data {
     VALUE subprocess_env;
     VALUE notes;
     VALUE finfo;
+    VALUE parsed_uri;
     VALUE attributes;
     VALUE error_message;
     VALUE exception;
@@ -85,7 +90,7 @@ static VALUE fname(VALUE self, VALUE val) \
     data = get_request_data(self); \
     data->request->member = \
 	apr_pstrndup(data->request->pool, \
-		    RSTRING(val)->ptr, RSTRING(val)->len); \
+		    RSTRING_PTR(val), RSTRING_LEN(val)); \
     return val; \
 }
 #else
@@ -97,7 +102,7 @@ static VALUE fname(VALUE self, VALUE val) \
     data = get_request_data(self); \
     data->request->member = \
 	ap_pstrndup(data->request->pool, \
-		    RSTRING(val)->ptr, RSTRING(val)->len); \
+		    RSTRING_PTR(val), RSTRING_LEN(val)); \
     return val; \
 }
 #endif
@@ -106,6 +111,10 @@ static VALUE fname(VALUE self, VALUE val) \
 	DEFINE_INT_ATTR_READER(fname, request_data, request->member)
 #define REQUEST_INT_ATTR_WRITER(fname, member) \
 	DEFINE_INT_ATTR_WRITER(fname, request_data, request->member)
+#define REQUEST_BOOL_ATTR_READER(fname, member) \
+	DEFINE_BOOL_ATTR_READER(fname, request_data, request->member)
+#define REQUEST_BOOL_ATTR_WRITER(fname, member) \
+	DEFINE_BOOL_ATTR_WRITER(fname, request_data, request->member)
 
 static void request_mark(request_data *data)
 {
@@ -119,6 +128,7 @@ static void request_mark(request_data *data)
     rb_gc_mark(data->subprocess_env);
     rb_gc_mark(data->notes);
     rb_gc_mark(data->finfo);
+    rb_gc_mark(data->parsed_uri);
     rb_gc_mark(data->attributes);
     rb_gc_mark(data->error_message);
     rb_gc_mark(data->exception);
@@ -182,6 +192,7 @@ static VALUE apache_request_new(request_rec *r)
     data->subprocess_env = Qnil;
     data->notes = Qnil;
     data->finfo = Qnil;
+    data->parsed_uri = Qnil;
     data->attributes = Qnil;
     data->error_message = Qnil;
     data->exception = Qnil;
@@ -192,20 +203,23 @@ static VALUE apache_request_new(request_rec *r)
     data->cookies = rb_hash_new();
     data->param_table = Qnil;
     data->options = rb_hash_new();
-    opts_arr = apr_table_elts(dconf->options);
-    opts = (table_entry *) opts_arr->elts;
-    for (i = 0; i < opts_arr->nelts; i++) {
-        if (opts[i].key == NULL)
-	    continue;
-	rb_hash_aset(data->options,
-		     rb_tainted_str_new2(opts[i].key),
-		     rb_tainted_str_new2(opts[i].val));
+    if (dconf) {
+        opts_arr = apr_table_elts(dconf->options);
+        opts = (table_entry *) opts_arr->elts;
+        for (i = 0; i < opts_arr->nelts; i++) {
+            if (opts[i].key == NULL)
+                continue;
+            rb_hash_aset(data->options,
+                         rb_tainted_str_new2(opts[i].key),
+                         rb_tainted_str_new2(opts[i].val));
+        }
     }
     
     rb_apache_register_object(obj);
     if (r->request_config) {
 	ruby_request_config *rconf = get_request_config(r);
-	rconf->request_object = obj;
+	if (rconf)
+	    rconf->request_object = obj;
     }
     apr_pool_cleanup_register(r->pool, (void *) r,
 			cleanup_request_object, apr_pool_cleanup_null);
@@ -231,7 +245,7 @@ VALUE rb_get_request_object(request_rec *r)
     if (r == NULL) return Qnil;
     if (r->request_config) {
 	ruby_request_config *rconf = get_request_config(r);
-	if (!NIL_P(rconf->request_object))
+	if (rconf && !NIL_P(rconf->request_object))
 	    return rconf->request_object;
     }
     return apache_request_new(r);
@@ -258,7 +272,7 @@ long rb_apache_request_length(VALUE self)
     request_data *data;
 
     data = get_request_data(self);
-    return RSTRING(data->outbuf)->len;
+    return RSTRING_LEN(data->outbuf);
 }
 
 static VALUE request_output_buffer(VALUE self)
@@ -282,8 +296,7 @@ static VALUE request_cancel(VALUE self)
     request_data *data;
 
     data = get_request_data(self);
-    RSTRING(data->outbuf)->len = 0;
-    RSTRING(data->outbuf)->ptr[0] = '\0';
+    rb_str_resize(data->outbuf, 0);
     return Qnil;
 }
 
@@ -343,12 +356,12 @@ static VALUE request_write(VALUE self, VALUE str)
 	if (data->request->header_only &&
 	    FL_TEST(self, REQ_SENT_HEADER))
 	    return INT2NUM(0);
-	len = ap_rwrite(RSTRING(str)->ptr, RSTRING(str)->len, data->request);
+	len = ap_rwrite(RSTRING_PTR(str), RSTRING_LEN(str), data->request);
 	ap_rflush(data->request);
     }
     else {
-	rb_str_cat(data->outbuf, RSTRING(str)->ptr, RSTRING(str)->len);
-	len = RSTRING(str)->len;
+	rb_str_cat(data->outbuf, RSTRING_PTR(str), RSTRING_LEN(str));
+	len = RSTRING_LEN(str);
     }
     return INT2NUM(len);
 }
@@ -413,7 +426,7 @@ static VALUE request_printf(int argc, VALUE *argv, VALUE out)
 
 static VALUE request_puts _((int, VALUE*, VALUE));
 
-#if RUBY_VERSION_CODE >= 190 && RUBY_RELEASE_CODE > 20050304
+#ifdef RUBY_VM
 static VALUE request_puts_ary(VALUE ary, VALUE out, int recur)
 #else
 static VALUE request_puts_ary(VALUE ary, VALUE out)
@@ -422,9 +435,9 @@ static VALUE request_puts_ary(VALUE ary, VALUE out)
     VALUE tmp;
     int i;
 
-    for (i=0; i<RARRAY(ary)->len; i++) {
-	tmp = RARRAY(ary)->ptr[i];
-#if RUBY_VERSION_CODE >= 190 && RUBY_RELEASE_CODE > 20050304
+    for (i=0; i<RARRAY_LEN(ary); i++) {
+	tmp = RARRAY_PTR(ary)[i];
+#ifdef RUBY_VM
 	if (recur) {
 #else
 	if (rb_inspecting_p(tmp)) {
@@ -452,7 +465,7 @@ static VALUE request_puts(int argc, VALUE *argv, VALUE out)
 	    line = STRING_LITERAL("nil");
 	    break;
 	  case T_ARRAY:
-#if RUBY_VERSION_CODE >= 190 && RUBY_RELEASE_CODE > 20050304
+#ifdef RUBY_VM
 	    rb_exec_recursive(request_puts_ary, argv[i], out);
 #else
 	    rb_protect_inspect(request_puts_ary, argv[i], out);
@@ -464,7 +477,7 @@ static VALUE request_puts(int argc, VALUE *argv, VALUE out)
 	}
 	line = rb_obj_as_string(line);
 	request_write(out, line);
-	if (RSTRING(line)->ptr[RSTRING(line)->len-1] != '\n') {
+	if (RSTRING_PTR(line)[RSTRING_LEN(line)-1] != '\n') {
 	    request_write(out, rb_default_rs);
 	}
     }
@@ -517,14 +530,14 @@ void rb_apache_request_flush(VALUE self)
     }
     if (data->request->header_only &&
 	FL_TEST(self, REQ_SENT_HEADER)) {
-	RSTRING(data->outbuf)->len = 0;
+	rb_str_resize(data->outbuf, 0);
 	return;
     }
-    if (RSTRING(data->outbuf)->len > 0) {
-	ap_rwrite(RSTRING(data->outbuf)->ptr, RSTRING(data->outbuf)->len,
+    if (RSTRING_LEN(data->outbuf) > 0) {
+	ap_rwrite(RSTRING_PTR(data->outbuf), RSTRING_LEN(data->outbuf),
 		  data->request);
 	ap_rflush(data->request);
-	RSTRING(data->outbuf)->len = 0;
+	rb_str_resize(data->outbuf, 0);
     }
 }
 
@@ -602,6 +615,8 @@ REQUEST_INT_ATTR_WRITER(request_set_status, status);
 REQUEST_STRING_ATTR_READER(request_get_status_line, status_line);
 REQUEST_STRING_ATTR_WRITER(request_set_status_line, status_line);
 REQUEST_STRING_ATTR_READER(request_the_request, the_request);
+REQUEST_BOOL_ATTR_READER(request_get_assbackwards, assbackwards);
+REQUEST_BOOL_ATTR_WRITER(request_set_assbackwards, assbackwards);
 REQUEST_STRING_ATTR_READER(request_request_method, method);
 REQUEST_INT_ATTR_READER(request_method_number, method_number);
 REQUEST_INT_ATTR_READER(request_get_allowed, allowed);
@@ -650,10 +665,9 @@ static VALUE request_set_content_type(VALUE self, VALUE str)
     }
     else {
 	Check_Type(str, T_STRING);
-	str = rb_funcall(str, rb_intern("downcase"), 0);
 	data->request->content_type =
 	    apr_pstrndup(data->request->pool,
-			RSTRING(str)->ptr, RSTRING(str)->len);
+			RSTRING_PTR(str), RSTRING_LEN(str));
     }
     return str;
 }
@@ -668,11 +682,10 @@ static VALUE request_set_content_encoding(VALUE self, VALUE str)
     }
     else {
 	Check_Type(str, T_STRING);
-	str = rb_funcall(str, rb_intern("downcase"), 0);
 	data->request->content_encoding =
 	    apr_pstrndup(data->request->pool,
-			RSTRING(str)->ptr,
-			RSTRING(str)->len);
+			RSTRING_PTR(str),
+			RSTRING_LEN(str));
     }
     return str;
 }
@@ -701,18 +714,17 @@ static VALUE request_set_content_languages(VALUE self, VALUE ary)
     }
     else {
 	Check_Type(ary, T_ARRAY);
-	for (i = 0; i < RARRAY(ary)->len; i++) {
-	    Check_Type(RARRAY(ary)->ptr[i], T_STRING);
+	for (i = 0; i < RARRAY_LEN(ary); i++) {
+	    Check_Type(RARRAY_PTR(ary)[i], T_STRING);
 	}
 	data->request->content_languages =
-	    apr_array_make(data->request->pool, RARRAY(ary)->len, sizeof(char *));
-	for (i = 0; i < RARRAY(ary)->len; i++) {
-	    VALUE str = RARRAY(ary)->ptr[i];
-	    str = rb_funcall(str, rb_intern("downcase"), 0);
+	    apr_array_make(data->request->pool, RARRAY_LEN(ary), sizeof(char *));
+	for (i = 0; i < RARRAY_LEN(ary); i++) {
+	    VALUE str = RARRAY_PTR(ary)[i];
 	    *(char **) apr_array_push(data->request->content_languages) =
 		apr_pstrndup(data->request->pool,
-			    RSTRING(str)->ptr,
-			    RSTRING(str)->len);
+			    RSTRING_PTR(str),
+			    RSTRING_LEN(str));
 	}
     }
     return ary;
@@ -876,6 +888,17 @@ static VALUE request_finfo(VALUE self)
     return data->finfo;
 }
 
+static VALUE request_parsed_uri(VALUE self)
+{
+    request_data *data;
+    
+    data = get_request_data(self);
+    if (NIL_P(data->parsed_uri)) {
+      data->parsed_uri = rb_apache_uri_new(&data->request->parsed_uri);
+    }
+    return data->parsed_uri;
+}
+
 static VALUE request_attributes(VALUE self)
 {
     request_data *data;
@@ -915,8 +938,16 @@ static VALUE request_get_client_block(VALUE self, VALUE length)
 
     data = get_request_data(self);
     len = NUM2INT(length);
+#if RUBY_VERSION_CODE < 180
+    {
+	char *buf = (char *) ap_palloc(data->request->pool, len);
+	len = ap_get_client_block(data->request, buf, len);
+	result = rb_tainted_str_new(buf, len);
+	return result;
+    }
+#else
     result = rb_str_buf_new(len);
-    len = ap_get_client_block(data->request, RSTRING(result)->ptr, len);
+    len = ap_get_client_block(data->request, RSTRING_PTR(result), len);
     switch (len) {
     case -1:
 	rb_raise(rb_eApachePrematureChunkEndError, "premature chunk end");
@@ -924,11 +955,11 @@ static VALUE request_get_client_block(VALUE self, VALUE length)
     case 0:
 	return Qnil;
     default:
-	RSTRING(result)->ptr[len] = '\0';
-	RSTRING(result)->len = len;
+	rb_str_resize(result, len);
 	OBJ_TAINT(result);
 	return result;
     }
+#endif
 }
 
 static VALUE read_client_block(request_rec *r, int len)
@@ -996,9 +1027,9 @@ static VALUE request_getc(VALUE self)
 
     data = get_request_data(self);
     str = read_client_block(data->request, 1);
-    if (NIL_P(str) || RSTRING(str)->len == 0)
+    if (NIL_P(str) || RSTRING_LEN(str) == 0)
 	return Qnil;
-    return INT2FIX(RSTRING(str)->ptr[0]);
+    return INT2FIX(RSTRING_PTR(str)[0]);
 }
 
 static VALUE request_eof(VALUE self)
@@ -1040,6 +1071,14 @@ static VALUE request_default_type(VALUE self)
     return type ? rb_tainted_str_new2(type) : Qnil;
 }
 
+static VALUE request_default_port(VALUE self)
+{
+    request_data *data;
+
+    data = get_request_data(self);
+    return INT2NUM(ap_default_port(data->request));
+}
+
 static VALUE request_remote_host(int argc, VALUE *argv, VALUE self)
 {
     request_data *data;
@@ -1190,7 +1229,7 @@ static VALUE request_hard_timeout(VALUE self, VALUE name)
 
     Check_Type(name, T_STRING);
     data = get_request_data(self);
-    s = ap_pstrndup(data->request->pool, RSTRING(name)->ptr, RSTRING(name)->len);
+    s = ap_pstrndup(data->request->pool, RSTRING_PTR(name), RSTRING_LEN(name));
     ap_hard_timeout(s, data->request);
 #endif
     return Qnil;
@@ -1203,7 +1242,7 @@ static VALUE request_soft_timeout(VALUE self, VALUE name)
 
     Check_Type(name, T_STRING);
     data = get_request_data(self);
-    s = apr_pstrndup(data->request->pool, RSTRING(name)->ptr, RSTRING(name)->len);
+    s = apr_pstrndup(data->request->pool, RSTRING_PTR(name), RSTRING_LEN(name));
     ap_soft_timeout(s, data->request);
     return Qnil;
 }
@@ -1351,8 +1390,8 @@ static VALUE request_set_user(VALUE self, VALUE val)
     Check_Type(val, T_STRING);
     data = get_request_data(self);
     data->request->user = apr_pstrndup(data->request->pool,
-				      RSTRING(val)->ptr,
-				      RSTRING(val)->len);
+				      RSTRING_PTR(val),
+				      RSTRING_LEN(val));
     return val;
 }
 #else /* Apache 1.x */
@@ -1445,8 +1484,8 @@ static VALUE request_set_auth_type(VALUE self, VALUE val)
     conf = (core_dir_config *)
 	ap_get_module_config(data->request->per_dir_config, &core_module);
     conf->ap_auth_type = apr_pstrndup(data->request->pool,
-				     RSTRING(val)->ptr,
-				     RSTRING(val)->len);
+				     RSTRING_PTR(val),
+				     RSTRING_LEN(val));
     ap_set_module_config(data->request->per_dir_config, &core_module, conf);
     return val;
 }
@@ -1477,8 +1516,8 @@ static VALUE request_set_auth_name(VALUE self, VALUE val)
     conf = (core_dir_config *)
 	ap_get_module_config(data->request->per_dir_config, &core_module);
     conf->ap_auth_name = apr_pstrndup(data->request->pool,
-				     RSTRING(val)->ptr,
-				     RSTRING(val)->len);
+				     RSTRING_PTR(val),
+				     RSTRING_LEN(val));
     ap_set_module_config(data->request->per_dir_config, &core_module, conf);
     return val;
 }
@@ -1513,7 +1552,7 @@ static VALUE request_bytes_sent(VALUE self)
 
 static VALUE request_send_fd(VALUE self, VALUE io)
 {
-    OpenFile *fptr;
+    rb_io_t *fptr;
     request_data *data;
 #ifdef APACHE2
     apr_size_t bytes_sent;
@@ -1533,7 +1572,7 @@ static VALUE request_send_fd(VALUE self, VALUE io)
     GetOpenFile(io, fptr);
 
 #ifdef APACHE2
-#if RUBY_VERSION_CODE >= 190
+#ifdef RUBY_VM
     fd = fptr->fd;
 #else
     fd = fileno(fptr->f);
@@ -1542,11 +1581,11 @@ static VALUE request_send_fd(VALUE self, VALUE io)
 	rb_raise(rb_eIOError, "apr_os_file_put() failed");
     }
     if (fstat(fd, &st) == -1) {
-	rb_sys_fail(fptr->path);
+	rb_sys_fail(IO_PATH(fptr));
     }
     ap_send_fd(file, data->request, 0, st.st_size, &bytes_sent);
 #else
-#if RUBY_VERSION_CODE >= 190
+#ifdef RUBY_VM
     f = rb_io_stdio_file(fptr);
 #else
     f = fptr->f;
@@ -1766,7 +1805,11 @@ static VALUE request_uploads( VALUE self )
 
     if ( !data->apreq->parsed ) rb_funcall( self, rb_intern("parse"), 0 );
 
+#if RUBY_VM
+    if (!RHASH(data->upload_table)->ntbl->num_entries) {
+#else
     if (!RHASH(data->upload_table)->tbl->num_entries) {
+#endif
 	for ( upload = ApacheRequest_upload(data->apreq);
 	      upload;
 	      upload = upload->next )
@@ -1890,12 +1933,12 @@ static VALUE request_set_parse_option( VALUE pair, VALUE self )
     VALUE optval;
 
     Check_Type( pair, T_ARRAY );
-    if ( !RARRAY(pair)->len == 2 )
+    if ( !RARRAY_LEN(pair) == 2 )
 	rb_raise( rb_eArgError, "Expected an array of 2 elements, not %d",
-		  RARRAY(pair)->len );
+		  (int) RARRAY_LEN(pair) );
 
-    opt = rb_to_id( *(RARRAY(pair)->ptr) );
-    optval = *(RARRAY(pair)->ptr + 1);
+    opt = rb_to_id( *(RARRAY_PTR(pair)) );
+    optval = *(RARRAY_PTR(pair) + 1);
 
     if ( opt == id_post_max ) {
 	request_post_max_eq( self, optval );
@@ -1913,8 +1956,9 @@ static VALUE request_set_parse_option( VALUE pair, VALUE self )
 	request_upload_hook_eq( self, optval );
     }
     else {
+	VALUE s = rb_inspect(*( RARRAY_PTR(pair) ));
 	rb_raise( rb_eArgError, "Unknown option %s",
-		  rb_inspect(*( RARRAY(pair)->ptr )) );
+		  StringValuePtr(s) );
     }
 
     return optval;
@@ -2036,7 +2080,11 @@ static VALUE request_cookies( VALUE self )
 
     if ( !data->apreq->parsed ) rb_funcall( self, rb_intern("parse"), 0 );
 
+#ifdef RUBY_VM
+    if ( ! RHASH(data->cookies)->ntbl->num_entries ) {
+#else
     if ( ! RHASH(data->cookies)->tbl->num_entries ) {
+#endif
 	ApacheCookieJar *jar = ApacheCookie_parse( data->request, NULL );
     int i;
 
@@ -2126,6 +2174,8 @@ void rb_init_apache_request()
     rb_define_method(rb_cApacheRequest, "allowed=", request_set_allowed, 1);
     rb_define_method(rb_cApacheRequest, "the_request",
 		     request_the_request, 0);
+    rb_define_method(rb_cApacheRequest, "assbackwards?", request_get_assbackwards, 0);
+    rb_define_method(rb_cApacheRequest, "assbackwards=", request_set_assbackwards, 1);
     rb_define_method(rb_cApacheRequest, "header_only?", request_header_only, 0);
     rb_define_method(rb_cApacheRequest, "args", request_get_args, 0);
     rb_define_method(rb_cApacheRequest, "args=", request_set_args, 1);
@@ -2151,6 +2201,7 @@ void rb_init_apache_request()
 		     request_subprocess_env, 0);
     rb_define_method(rb_cApacheRequest, "notes", request_notes, 0);
     rb_define_method(rb_cApacheRequest, "finfo", request_finfo, 0);
+    rb_define_method(rb_cApacheRequest, "parsed_uri", request_parsed_uri, 0);
     rb_define_method(rb_cApacheRequest, "attributes", request_attributes, 0);
     rb_define_method(rb_cApacheRequest, "setup_client_block",
 		     request_setup_client_block, -1);
@@ -2171,6 +2222,7 @@ void rb_init_apache_request()
 		     request_allow_overrides, 0);
     rb_define_method(rb_cApacheRequest, "default_type",
 		     request_default_type, 0);
+    rb_define_method(rb_cApacheRequest, "default_port", request_default_port, 0);
     rb_define_method(rb_cApacheRequest, "remote_host",
 		     request_remote_host, -1);
     rb_define_method(rb_cApacheRequest, "remote_logname",
